// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

import 'dart:mirrors';
import 'package:constrain/metadata.dart';
import 'package:constrain/src/default_validator.dart';
import 'package:constrain/constraint.dart';
import 'package:constrain/validator.dart';
import 'package:either/either.dart';
import 'dart:async';


// Highly experimental and untested - don't use


class ConstrainedFunctionProxy {
  final Validator validator;

  ConstrainedFunctionProxy(this.validator);

  invoke(ClosureMirror cm,
      List positionalParameters,
      Map<Symbol, dynamic> namedArguments,
      { Iterable<ConstraintGroup> groups,
        bool validateParameters: true, bool validateReturn: false }) {


    final violations = validator.validateFunctionParameters(cm.reflectee,
        positionalParameters, namedParameters: namedArguments, groups: groups);

    if (violations.isNotEmpty) {
      throw new ConstraintViolationException(violations);
    }

    final result = cm.apply(positionalParameters, namedArguments);

    // don't want to turn a non future into a future
    return result is Future ? result.then((r) =>
        _handleReturn(cm, validateReturn, r, groups))
        : _handleReturn(cm, validateReturn, result, groups);

  }

  _handleReturn(ClosureMirror cm, bool validateReturn, result, groups) {
    if (validateReturn) {
      final Set<ConstraintViolation> returnViolations =
          validator.validateFunctionReturn(cm.reflectee, result, groups: groups);

      if (returnViolations.isNotEmpty) {
        throw new ConstraintViolationException(returnViolations);
      }
    }
    return result;

  }
}

class Foo {
  String blah(int bar) => "";
}

@proxy class FooProxy extends MethodValidatingProxy implements Foo {
  FooProxy(ConstrainedFunctionProxy functionProxy, Foo delegate)
      : super(functionProxy, delegate);

  noSuchMethod(i) => super.noSuchMethod(i);
}

class MethodValidatingProxy  {
  final ConstrainedFunctionProxy _functionProxy;
  final delegate;
  final InstanceMirror _delegateMirror;

  MethodValidatingProxy(this._functionProxy, delegate)
      : this.delegate = delegate,
        _delegateMirror = reflect(delegate);

  noSuchMethod(Invocation invocation) {
//    final method = _delegateMirror.type.declarations[invocation.memberName];
    final method = _delegateMirror.getField(invocation.memberName);

    if (method == null) {
      throw new NoSuchMethodError(delegate, invocation.memberName,
          invocation.positionalArguments, invocation.namedArguments);
    }

    return _functionProxy.invoke(method, invocation.positionalArguments,
        invocation.namedArguments);
  }
}



